# -------------------------------------------------------------------------------------------------
# test_custom_mp_xi2.py: Customizing managers andd Proxies in Multiprocessing                              
#
# Theory:      Multiprocessing and Cutomizing using Inheritance 
#
# Description: Synchronization types like locks, conditions and queues are handy in multiprocessing
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double
import time
import threading
import gc
import random

import operator

import queue # <--- NOT Queue as in the original example
#from Queue import Empty <--- changed
from queue import Empty

# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
import multiprocessing

from multiprocessing.process import Process
from multiprocessing.queues import Queue
from multiprocessing.connection import Pipe
from multiprocessing.synchronize import Lock
from multiprocessing.pool import Pool
from multiprocessing.sharedctypes import Value, Array
from multiprocessing.managers import BaseManager, BaseProxy
from multiprocessing import Manager

from multiprocessing import current_process
from multiprocessing import freeze_support

# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from pool2xi import Pool2Xi
from value2xi import Value2Xi
from array2xi import Array2Xi
from basemanager2xi import BaseManager2Xi
from baseproxy2xi import BaseProxy2Xi
from manager2xi import Manager2Xi


# -------------------------------------------------------------------------------------------------
# Foo                              
# -------------------------------------------------------------------------------------------------
#
class Foo(object):
    # ---------------------------------------------------------------------------------------------
    # f method
    # ---------------------------------------------------------------------------------------------
    def f(self):
        print ('you called Foo.f()')
    # ---------------------------------------------------------------------------------------------
    # g method
    # --------------------------------------------------------------------------------------------    
    def g(self):
        print ('you called Foo.g()')
    # ---------------------------------------------------------------------------------------------
    # _h method
    # -------------------------------------------------------------------------------------------
    def _h(self):
        print ('you called Foo._h()')

# -------------------------------------------------------------------------------------------------
# Dragon                              
# -------------------------------------------------------------------------------------------------
#
class Dragon(object):
    # ---------------------------------------------------------------------------------------------
    # f method
    # ---------------------------------------------------------------------------------------------
    def f(self):
        print ('you called Dragon.f()')
    # ---------------------------------------------------------------------------------------------
    # g method
    # --------------------------------------------------------------------------------------------    
    def g(self):
        print ('you called Dragon.g()')
    # ---------------------------------------------------------------------------------------------
    # _h method
    # -------------------------------------------------------------------------------------------
    def _h(self):
        print ('you called Dragon._h()')        
        
# -------------------------------------------------------------------------------------------------
# A simple generator function
def baz():
    for i in range(10):
        yield i*i

# -------------------------------------------------------------------------------------------------
# GeneratorProxy                              
#                 Proxy type for generator objects
# -------------------------------------------------------------------------------------------------
class GeneratorProxy(BaseProxy):
    _exposed_ = ('next', '__next__')
    def __iter__(self):
        return self
    def next(self):
        return self._callmethod('next')
    def __next__(self):
        return self._callmethod('__next__')

# -------------------------------------------------------------------------------------------------
# GeneratorProxy2Xi                              
#                 Proxy type for generator objects
# -------------------------------------------------------------------------------------------------
class GeneratorProxy2Xi(BaseProxy2Xi):
    _exposed_ = ('next', '__next__')
    def __iter__(self):
        return self
    def next(self):
        return self._callmethod('next')
    def __next__(self):
        return self._callmethod('__next__')        
                
# -------------------------------------------------------------------------------------------------
# Function to return the operator module
def get_operator_module():
    return operator

# -------------------------------------------------------------------------------------------------
# MyManager                              
# -------------------------------------------------------------------------------------------------
#
class MyManager(BaseManager):
    pass

# -------------------------------------------------------------------------------------------------
#
# register the Foo class; make `f()` and `g()` accessible via proxy
MyManager.register('Foo1', Foo)

# register the Foo class; make `g()` and `_h()` accessible via proxy
MyManager.register('Foo2', Foo, exposed=('g', '_h'))

# register the generator function baz; use `GeneratorProxy` to make proxies
MyManager.register('baz', baz, proxytype=GeneratorProxy)

# register get_operator_module(); make public functions accessible via proxy
MyManager.register('operator', get_operator_module)

# -------------------------------------------------------------------------------------------------
# phase1: standard multiprocessing
# -------------------------------------------------------------------------------------------------
#
def phase1():
    freeze_support()

    manager = MyManager()
    manager.start()

    print ('-' * 20)

    f1 = manager.Foo1()
    f1.f()
    f1.g()
    assert not hasattr(f1, '_h')
    assert sorted(f1._exposed_) == sorted(['f', 'g'])

    print ('-' * 20)

    f2 = manager.Foo2()
    f2.g()
    f2._h()
    assert not hasattr(f2, 'f')
    assert sorted(f2._exposed_) == sorted(['g', '_h'])

    print ('-' * 20)

    it = manager.baz()
    for i in it:
        print ('<%d>' % i,)
    print

    print ('-' * 20)

    op = manager.operator()
    print ('op.add(23, 45) =', op.add(23, 45))
    print ('op.pow(2, 94) =', op.pow(2, 94))
    #print ('op.getslice(range(10), 2, 6) =', op.getslice(range(10), 2, 6)) <--- error: no attribute such as getslice !
    #print ('op.repeat(range(5), 3) =', op.repeat(range(5), 3)) <--- error: range can not be repeated
    print ('op._exposed_ =', op._exposed_)



# -------------------------------------------------------------------------------------------------
# MyManager2Xi                              
# -------------------------------------------------------------------------------------------------
#
class MyManager2Xi(BaseManager2Xi):
    pass




# -------------------------------------------------------------------------------------------------
# phase2: dragonized multiprocessing
# -------------------------------------------------------------------------------------------------
def phase2():

    # ---------------------------------------------------------------------------------------------
    # registrations
    # ---------------------------------------------------------------------------------------------
    # register Dragon class as DragonXi1; make `f()` and `g()` accessible via proxy
    MyManager2Xi.register('DragonXi1', Dragon)
    
    # register Dragon class as DragonXi2; make `g()` and `_h()` accessible via proxy
    MyManager2Xi.register('DragonXi2', Dragon, exposed=('g', '_h'))
    
    # register the generator function baz as sense2xi; use customized `GeneratorProxy2Xi` to make proxies
    MyManager2Xi.register('sense2xi', baz, proxytype=GeneratorProxy2Xi)
    
    # register get_operator_module as operator2xi; make public functions accessible via proxy
    MyManager2Xi.register('operator2xi', get_operator_module)
    
    # ---------------------------------------------------------------------------------------------
    # object manager2xi created from MyManager2Xi
    # ---------------------------------------------------------------------------------------------    
    print ('=' * 20)
    print("\n Dragonized multiprocessing")
    print("\n object manager2xi created from MyManager2Xi")
    manager2xi = MyManager2Xi()
    manager2xi.start()

    # ---------------------------------------------------------------------------------------------
    # class registered with name DragonXi1 is used
    # ---------------------------------------------------------------------------------------------   
    print ('-' * 20)
    print("\n class registered with name DragonXi1 is used")
    f12xi = manager2xi.DragonXi1()
    f12xi.f()
    f12xi.g()
    assert not hasattr(f12xi, '_h')
    assert sorted(f12xi._exposed_) == sorted(['f', 'g'])

    # ---------------------------------------------------------------------------------------------
    # class registered with name DragonXi2 is used
    # ---------------------------------------------------------------------------------------------
    print ('-' * 20)
    print("\n class registered with name DragonXi2 is used")
    f22xi = manager2xi.DragonXi2()
    f22xi.g()
    f22xi._h()
    assert not hasattr(f22xi, 'f')
    assert sorted(f22xi._exposed_) == sorted(['g', '_h'])

    # ---------------------------------------------------------------------------------------------
    # function registered with name sense2xi is called
    # ---------------------------------------------------------------------------------------------
    print ('-' * 20)
    print("\n function registered with name sense2xi is called")
    it2xi = manager2xi.sense2xi()
    for i in it2xi:
        print ('<%d>' % i,)
    print

    # ---------------------------------------------------------------------------------------------
    # operator registered with name operator2xi and it's operations are used
    # ---------------------------------------------------------------------------------------------
    print ('-' * 20)
    print("\n operator registered with name operator2xi and it's operations are used")
    op2xi = manager2xi.operator2xi()
    print ('op2xi.add(23, 45) =', op2xi.add(23, 45))
    print ('op2xi.pow(2, 94) =', op2xi.pow(2, 94))
    #print ('op.getslice(range(10), 2, 6) =', op.getslice(range(10), 2, 6)) <--- error: no attribute such as getslice !
    #print ('op.repeat(range(5), 3) =', op.repeat(range(5), 3)) <--- error: range can not be repeated
    print ('op2xi._exposed_ =', op2xi._exposed_)


# -------------------------------------------------------------------------------------------------
# main                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    
    phase1()   

    phase2()

# -------------------------------------------------------------------------------------------------
# test outputs                               
# -------------------------------------------------------------------------------------------------    
# --------------------
# <0>
# <1>
# <4>
# <9>
# <16>
# <25>
# <36>
# <49>
# <64>
# <81>
# --------------------
# op.add(23, 45) = 68
# op.pow(2, 94) = 19807040628566084398385987584
# op._exposed_ = ('abs', 'add', 'and_', 'attrgetter', 'concat', 'contains', 'countOf', 'delitem', 'eq', 'floordiv', 'ge', 'getitem', 'gt', 'iadd', 'iand', 'iconcat', 'ifloordiv', 'ilshift', 'imod', 'imul', 'index', 'indexOf', 'inv', 'invert', 'ior', 'ipow', 'irepeat', 'irshift', 'isMappingType', 'isNumberType', 'isSequenceType', 'is_', 'is_not', 'isub', 'itemgetter', 'itruediv', 'ixor', 'le', 'lshift', 'lt', 'methodcaller', 'mod', 'mul', 'ne', 'neg', 'not_', 'or_', 'pos', 'pow', 'repeat', 'rshift', 'setitem', 'sub', 'truediv', 'truth', 'xor')
# ====================
# 
#  Dragonized multiprocessing
# 
#  object manager2xi created from MyManager2Xi
# --------------------
# 
#  class registered with name DragonXi1 is used
# --------------------
# 
#  class registered with name DragonXi2 is used
# --------------------
# 
#  function registered with name sense2xi is called
# <0>
# <1>
# <4>
# <9>
# <16>
# <25>
# <36>
# <49>
# <64>
# <81>
# --------------------
# 
# operator registered with name operator2xi and it's operations are used
# op2xi.add(23, 45) = 68
# op2xi.pow(2, 94) = 19807040628566084398385987584
# op2xi._exposed_ = ('abs', 'add', 'and_', 'attrgetter', 'concat', 'contains', 'countOf', 'delitem', 'eq', 'floordiv', 'ge', 'getitem', 'gt', 'iadd', 'iand', 'iconcat', 'ifloordiv', 'ilshift', 'imod', 'imul', 'index', 'indexOf', 'inv', 'invert', 'ior', 'ipow', 'irepeat', 'irshift', 'isMappingType', 'isNumberType', 'isSequenceType', 'is_', 'is_not', 'isub', 'itemgetter', 'itruediv', 'ixor', 'le', 'lshift', 'lt', 'methodcaller', 'mod', 'mul', 'ne', 'neg', 'not_', 'or_', 'pos', 'pow', 'repeat', 'rshift', 'setitem', 'sub', 'truediv', 'truth', 'xor')
# 
# 
# -------------------------------------------------------------------------------------------------
# End Of test_custom_mp_xi2.py                               
# -------------------------------------------------------------------------------------------------